#define DONT_USE_HPGCC_LIBC
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include "../include/_hpapine_version.h"	/* hpapine_print_version() */
#include "../include/hpapine_gui.h"
#include "gui_x11.h"

const char *gui_name = "XLib (X11)";


static Display *dpy;
static int screen;
static Window win;
static GC gc;
static XColor xcolors[16];


static unsigned int winw;
static unsigned int winh;
static unsigned int winzoom = DEFAULT_WINZOOM;

static int is_gui_init = 0;

static void *gui_eventloop(void *arg);
static void *gui_screendriver(void *arg);
static void gui_close(void);

static const long eventMask = FocusChangeMask | KeyPressMask | KeyReleaseMask;



extern void gui_run(void)
{
    if (is_gui_init)
	return;
    is_gui_init = 1;

    hpapine_print_version();

    /* we use multithreading */
    if (!XInitThreads()) {
	fprintf(stderr,
		"XInitThreads() fails. Still running. Expect me to crash soon.\n");
    }

    /* Create and raise window */

    char *display_name = getenv("DISPLAY");
    unsigned long fore, back;

    winw = GUI_WIN_W * winzoom;
    winh = GUI_WIN_H * winzoom;

    dpy = XOpenDisplay(display_name);
    if (dpy == NULL) {
	fprintf(stderr, "cannot connect to X server '%s'\n", display_name);
	exit(1);
    }
    screen = DefaultScreen(dpy);

    /* Colors */
    int i;
    Colormap cmap = DefaultColormap(dpy, screen);

    for (i = 0; i < 16; i++) {
      XColor exact;
      char col[13];
      snprintf(col, 13, "rgb:%02x/%02x/%02x", i*255/15, i*255/15, i*255/15);
      if (! XAllocNamedColor (dpy, cmap, col, &xcolors[i], &exact)) {
	fprintf(stderr, "%s: cant't alloc color %s\n", __FUNCTION__, col);
	exit (1);
      }
    }

    back = WhitePixel(dpy, screen);
    fore = BlackPixel(dpy, screen);

    /**/

    win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
			      winw, winh, 5, fore, back);
    XSetStandardProperties(dpy, win, "hpapine", "hpapine", None, NULL, 0,
			   NULL);
    XSelectInput(dpy, win, eventMask);
    gc = XCreateGC(dpy, win, 0, 0);


    XClearWindow(dpy, win);
    XMapRaised(dpy, win);


    /* Now, initialize the threads */

    int tid;
    pthread_t thread_eventloop, thread_screen;
    pthread_attr_t attr;

    /* Launch the eventloop thread */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    tid = pthread_create(&thread_eventloop, &attr, gui_eventloop, NULL);
    if (tid) {
	perror("pthread_create");
	exit(1);
    }

    /* Launch the screen thread */
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    tid = pthread_create(&thread_screen, &attr, gui_screendriver, NULL);
    if (tid) {
	perror("pthread_create");
	exit(1);
    }
    pthread_attr_destroy(&attr);

    /* */
}

static void gui_close(void)
{
    /* kill threads */

    XLockDisplay(dpy);
    is_gui_init = 0;
    XFreeGC(dpy, gc);
    XDestroyWindow(dpy, win);
    XCloseDisplay(dpy);
    XUnlockDisplay(dpy);
}


static void *gui_eventloop(void *arg)
{
    XEvent event, event2;
    int ret;
    KeySym key;
    int index;

    while (1) {
	ret = 0;
	while (1) {
	    usleep(7000);	/* 0.007s */
	    XLockDisplay(dpy);
	    ret = XCheckMaskEvent(dpy, eventMask, &event);
	    XUnlockDisplay(dpy);
	    if (ret)
	      break;
	}


        index = (event.xkey.state & ShiftMask) ? 1 : 0;

	if (event.type == KeyPress) {
	    XLockDisplay(dpy);
	    key = XLookupKeysym(&event.xkey, index);
	    XUnlockDisplay(dpy);
	    keysym_press(key);
	} else if (event.type == KeyRelease) {
	    /* We check whether the XEvent is a user key release event
	     * or a auto-repeat generated event.
	     */
	    XLockDisplay(dpy);
	    if (XCheckTypedEvent(dpy, KeyPress, &event2)) {
		if (event2.xkey.keycode == event.xkey.keycode) {
		    XUnlockDisplay(dpy);
		    /*
		    fprintf(stderr, "Autorepeat (delta=%u)\n",
				event2.xkey.time - event.xkey.time);
		    */
		    continue;
		}
		else {
		    /* How could we put back the event at the place it was? */
		    /* XPutBackEvent(dpy, &event2); */

		    index = (event2.xkey.state & ShiftMask) ? 1 : 0;
		    KeySym key2 = XLookupKeysym(&event.xkey, index);
		    char *ks = XKeysymToString(key2);
		    fprintf(stderr, "Dropping `%s' press.\n", ks);
		}
	    }

	    key = XLookupKeysym(&event.xkey, index);
	    keysym_release(key);

	    XUnlockDisplay(dpy);
	} else if (event.type == FocusOut || event.type == FocusIn) {
	    keysym_releaseall();
	} else {
	}
    }

    gui_close();
    pthread_exit(NULL);
}

static void *gui_screendriver(void *arg)
{
    Pixmap db;			/* double buffer */
    int i, j;
    int c;

    XLockDisplay(dpy);
    XSetBackground(dpy, gc, xcolors[0xf].pixel);
    XSetForeground(dpy, gc, xcolors[0x0].pixel);

    db = XCreatePixmap(dpy, DefaultRootWindow(dpy), winw, winh,
		       DefaultDepth(dpy, screen));
    XUnlockDisplay(dpy);

    while (1) {
	XLockDisplay(dpy);

	/* build the pixmap */
	/* XFillRectangle(dpy, db, gc, 0, 0, winw, winh); */

	for (i = 0; i < 131; i++) {
	    for (j = 0; j < 80; j++) {
		c = screen_getpixlum(i, j);
		XSetForeground(dpy, gc, xcolors[c].pixel);
		XFillRectangle(dpy, db, gc, winzoom * i, winzoom * j,
				   winzoom, winzoom);
	    }
	}

	/* display the pixmap */
	XCopyArea(dpy, db, win, gc, 0, 0, winw, winh, 0, 0);
	XFlush(dpy);

	XUnlockDisplay(dpy);

	usleep(40000);		/* 0.04 s */
    }
    pthread_exit(NULL);
}

void gui_change_screenbuf(unsigned char *new)
{
    if ((unsigned long) new & 0x3)
	fprintf(stderr, "%s: using a non-aligned screen buffer...\n",
		__FUNCTION__);

    XLockDisplay(dpy);
    screen_buf = new;
    XUnlockDisplay(dpy);
}


void gui_bell(int tone, int duration)
{
    XKeyboardControl c;
    XKeyboardState oldstate;

    c.bell_pitch = tone;
    c.bell_duration = duration;

    gui_run();

    XLockDisplay(dpy);
    XGetKeyboardControl(dpy, &oldstate);

    XChangeKeyboardControl(dpy, KBBellPitch | KBBellDuration, &c);
    XBell(dpy, 0);

    /* restore old values */
    c.bell_pitch = oldstate.bell_pitch;
    c.bell_duration = oldstate.bell_duration;
    XChangeKeyboardControl(dpy, KBBellPitch | KBBellDuration, &c);
    XUnlockDisplay(dpy);
}
