#include <stddef.h>
#include "hpgraphics.h"
#include "hpgpriv.h"


void _hpg_draw(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    /*
     * Clip to clipping region.  Any drawing that lies entirely off the
     * screen, in either the x or y dimensions, is culled here.
     */
    if ((y < g->clipy1) || (y >= g->clipy2)) return;
    if (x < g->clipx1)
    {
        len += x - g->clipx1;
        x = g->clipx1;
    }
    if (x + len >= g->clipx2)
    {
        len = g->clipx2 - x;
    }

    if (len <= 0) return;

    /*
     * Delegate to mode-specific code for the remainder of the drawing.
     */
    (*g->drawfn)(g, x, y, len, patx, paty);
}


/* gray16 drawing */

/* The following cases are considered separately:
 *
 * - No pattern
 * - Pattern becomes the general case
 */

static void _gray16_draw_nonpat(hpg_t *g, int x, int y, int len)
{
    unsigned char color = g->color >> 4;
    unsigned char colormask = color | (color << 4);

    char *fb = g->drawfb + g->fbwidth * y + (x >> 1);

    if (((x + len - 1) ^ x) < 2)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = ((1 << (len << 2)) - 1) << ((x & 1) << 2);
        if (g->mode == HPG_MODE_PAINT) *fb &= ~(mask);
        *fb ^= mask & colormask;
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 1);

        if (x & 1)
        {
            int beginlen = 2 - (x & 1);
            unsigned char mask = ((1 << (beginlen << 2)) - 1) << ((x & 1) << 2);
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *(fb++) ^= mask & colormask;
        }

        while (fb < fbend)
        {
            if (g->mode == HPG_MODE_PAINT) *fb = colormask;
            else *fb ^= colormask;
            fb++;
        }

        int endlen = ((len + (x & 1) - 1) & 1) + 1;
        unsigned char mask = (1 << (endlen << 2)) - 1;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & colormask;
    }
}

static void _gray16_draw_pattern(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (patx == -1) _gray16_draw_nonpat(g, x, y, len);

    unsigned char color = g->color >> 4;
    unsigned char colormask = color | (color << 4);

    /*
     * Read the line from the pattern;
     */
    struct hpg_pattern *pat = g->pattern;

    if (pat->fixed)
    {
        patx = 0;
        paty = 0;
    }

    int idx;
    unsigned char patline = 0;

    idx = y - paty;
    while (idx < 0) idx += pat->height;
    while (idx >= pat->height) idx -= pat->height;
    patline = pat->buffer[idx];

    /*
     * Build the master pattern.  This is yet another bitmask applied to all
     * drawing operations, and it allows the user to draw a pattern onto the
     * screen.
     *
     * This is also a candidate to be cached in the graphics context.
     */
    unsigned int masterpattern = 0x00;
    unsigned char srcmask = 0x01;
    unsigned int dstmask = 0x0F;

    while (dstmask != 0)
    {
        if ((patline & srcmask) > 0)
        {
            masterpattern |= dstmask;
        }

        dstmask <<= 4;
        srcmask <<= 1;

        if (srcmask == 0) srcmask = 1;
    }

    idx = ((x & ~1) - patx) & 7;
    masterpattern = (masterpattern >> (idx << 2))
        | (masterpattern << (32 - (idx << 2)));

    char *fb = g->drawfb + g->fbwidth * y + (x >> 1);

    if (((x + len - 1) ^ x) < 2)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = ((1 << (len << 2)) - 1) << ((x & 1) << 2);
        mask &= (unsigned char) masterpattern;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~(mask);
        *fb ^= mask & colormask;
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 1);

        if (x & 1)
        {
            int beginlen = 2 - (x & 1);
            unsigned char mask = ((1 << (beginlen << 2)) - 1) << ((x & 1) << 2);
            mask &= (unsigned char) masterpattern;
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *(fb++) ^= mask & colormask;

            masterpattern = (masterpattern >> 8) | (masterpattern << 24);
        }

        while (fb < fbend)
        {
            unsigned char mask = (unsigned char) masterpattern;
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *fb ^= mask & colormask;
            fb++;

            masterpattern = (masterpattern >> 8) | (masterpattern << 24);
        }

        int endlen = ((len + (x & 1) - 1) & 1) + 1;
        unsigned char mask = (1 << (endlen << 2)) - 1;
        mask &= (unsigned char) masterpattern;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & colormask;
    }
}

/*
 * Select a drawing routine.  Many cases use special routines to
 * optimize the drawing process.  In the future, once accessors and
 * mutators are implemented for graphics contexts, this series of
 * conditionals should be changed to replacement of the drawfn function
 * pointer to add speed.
 */
void _hpg_gray16_draw(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (g->pattern == NULL)
    {
        _gray16_draw_nonpat(g, x, y, len);
    }
    else
    {
        _gray16_draw_pattern(g, x, y, len, patx, paty);
    }
}



/* gray4 drawing */


static void _gray4_draw_nonpat(hpg_t *g, int x, int y, int len)
{
    unsigned char color = g->color >> (6);
    unsigned char colormask =
        color | (color << 2) | (color << 4) | (color << 6);

    char *fb = g->drawfb + g->fbwidth * y + (x >> 2);

    if (((x + len - 1) ^ x) < 4)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = ((1 << (len << 1)) - 1) << ((x & 3) << 1);
        if (g->mode == HPG_MODE_PAINT) *fb &= ~(mask);
        *fb ^= mask & colormask;
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 2);

        if (x & 7)
        {
            int beginlen = 8 - (x & 3);
            unsigned char mask = ((1 << (beginlen << 1)) - 1) << ((x & 3) << 1);
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *(fb++) ^= mask & colormask;
        }

        while (fb < fbend)
        {
            if (g->mode == HPG_MODE_PAINT) *fb = colormask;
            else *fb ^= colormask;
            fb++;
        }

        int endlen = ((len + (x & 3) - 1) & 3) + 1;
        unsigned char mask = (1 << (endlen << 1)) - 1;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & colormask;
    }
}

static void _gray4_draw_pattern(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (patx == -1) _gray4_draw_nonpat(g, x, y, len);

    unsigned char color = g->color >> (6);
    unsigned char colormask =
        color | (color << 2) | (color << 4) | (color << 6);

    /*
     * Read the line from the pattern;
     */
    struct hpg_pattern *pat = g->pattern;

    if (pat->fixed)
    {
        patx = 0;
        paty = 0;
    }

    int idx;
    unsigned char patline = 0;

    idx = y - paty;
    while (idx < 0) idx += pat->height;
    while (idx >= pat->height) idx -= pat->height;
    patline = pat->buffer[idx];

    /*
     * Build the master pattern.  This is yet another bitmask applied to all
     * drawing operations, and it allows the user to draw a pattern onto the
     * screen.
     *
     * This is also a candidate to be cached in the graphics context.
     */
    unsigned short masterpattern = 0x00;
    unsigned char srcmask = 0x01;
    unsigned short dstmask = 0x03;

    while (dstmask != 0)
    {
        if ((patline & srcmask) > 0)
        {
            masterpattern |= dstmask;
        }

        dstmask <<= 2;
        srcmask <<= 1;

        if (srcmask == 0) srcmask = 1;
    }

    idx = ((x & ~3) - patx) & 7;
    masterpattern = (masterpattern >> (idx << 1))
        | (masterpattern << (16 - (idx << 1)));

    char *fb = g->drawfb + g->fbwidth * y + (x >> 2);

    if (((x + len - 1) ^ x) < 4)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = ((1 << (len << 1)) - 1) << ((x & 3) << 1);
        mask &= (unsigned char) masterpattern;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~(mask);
        *fb ^= mask & colormask;
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 2);

        if (x & 3)
        {
            int beginlen = 4 - (x & 3);
            unsigned char mask = ((1 << (beginlen << 1)) - 1) << ((x & 3) << 1);
            mask &= (unsigned char) masterpattern;
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *(fb++) ^= mask & colormask;

            masterpattern = (masterpattern >> 8) | (masterpattern << 8);
        }

        while (fb < fbend)
        {
            unsigned char mask = (unsigned char) masterpattern;
            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *fb ^= mask & colormask;
            fb++;

            masterpattern = (masterpattern >> 8) | (masterpattern << 8);
        }

        int endlen = ((len + (x & 3) - 1) & 3) + 1;
        unsigned char mask = ((1 << (endlen << 1)) - 1);
        mask &= (unsigned char) masterpattern;
        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & colormask;
    }
}

/*
 * Select a drawing routine.  Many cases use special routines to
 * optimize the drawing process.  In the future, once accessors and
 * mutators are implemented for graphics contexts, this series of
 * conditionals should be changed to replacement of the drawfn function
 * pointer to add speed.
 */
void _hpg_gray4_draw(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (g->pattern == NULL)
    {
        _gray4_draw_nonpat(g, x, y, len);
    }
    else
    {
        _gray4_draw_pattern(g, x, y, len, patx, paty);
    }
}


/* mono drawing */


static void _mono_draw_nonpat(hpg_t *g, int x, int y, int len)
{
    unsigned char colormask = (g->color >> 7) ? 0xFF : 0x00;
    unsigned char xormask = (g->mode == HPG_MODE_XOR) ? 0xFF : 0x00;
    char *fb = g->drawfb + g->fbwidth * y + (x >> 3);

    if (((x + len - 1) ^ x) < 8)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = ((1 << len) - 1) << (x & 7);
        *fb = (*fb & (~mask | xormask)) ^ (mask & colormask);
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 3);

        if (x & 7)
        {
            int beginlen = 8 - (x & 7);
            unsigned char mask = ((1 << beginlen) - 1) << (x & 7);
            *fb = (*fb & (~mask | xormask)) ^ (mask & colormask);
            fb++;
        }

        while (fb < fbend)
        {
            *(fb) = (*fb & xormask) ^ colormask;
            fb++;
        }

        int endlen = ((len + (x & 7) - 1) & 7) + 1;
        unsigned char mask = (1 << endlen) - 1;
        *fb = (*fb & (~mask | xormask)) ^ (mask & colormask);
    }
}

static void _mono_draw_pattern(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (patx == -1) _mono_draw_nonpat(g, x, y, len);

    unsigned char color = (g->color & 0x80) ? 0xFF : 0x00;

    /*
     * Build the master pattern.  This is yet another bitmask applied to all
     * drawing operations, and it allows the user to draw a pattern onto the
     * screen.
     *
     * This is also a candidate to be cached in the graphics context.
     */
    struct hpg_pattern *pat = g->pattern;
    if (pat->fixed) patx = paty = 0;

    int idx;

    idx = y - paty;
    while (idx < 0) idx += pat->height;
    while (idx >= pat->height) idx -= pat->height;
    int line = pat->buffer[idx];

    idx = (x - patx + (x & 7)) & 7;
    line = (line << (idx)) | (line >> (8 - idx));

    char *fb = g->drawfb + g->fbwidth * y + (x >> 3);

    if (((x + len - 1) ^ x) < 8)
    {
        /*
         * Single-byte section.  Handle specially.
         */
        unsigned char mask = (((1 << len) - 1) << (x & 7)) & line;

        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & color;
    }
    else
    {
        char *fbend = g->drawfb + g->fbwidth * y
            + ((x + len - 1) >> 3);

        /*  first */
        if ((x & 7) != 0)
        {
            int firstlen = 8 - (x & 7);
            unsigned char mask = (((1 << firstlen) - 1) << (x & 7)) & line;

            if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
            *fb ^= mask & color;
            fb++;
        }

        /*  middle */
        while (fb < fbend)
        {
            if (g->mode == HPG_MODE_PAINT) *fb &= ~line;
            *fb ^= line & color;
            fb++;
        }

        /*  end */
        int endlen = ((len + (x & 7) - 1) & 7) + 1;
        unsigned char mask = ((1 << endlen) - 1) & line;

        if (g->mode == HPG_MODE_PAINT) *fb &= ~mask;
        *fb ^= mask & color;
    }
}

/*
 * Select a drawing routine.  Many cases use special routines to
 * optimize the drawing process.  In the future, once accessors and
 * mutators are implemented for graphics contexts, this series of
 * conditionals should be changed to replacement of the drawfn function
 * pointer to add speed.
 */
void _hpg_mono_draw(hpg_t *g, int x, int y, int len, int patx, int paty)
{
    if (g->pattern == NULL)
    {
        _mono_draw_nonpat(g, x, y, len);
    }
    else
    {
        _mono_draw_pattern(g, x, y, len, patx, paty);
    }
}
