#include <hpstdlib.h>
#include <hpstring.h>
#include "hpgraphics.h"
#include "hpgpriv.h"

#include <ctype.h>


/*
 * Convert an RGB color value to grayscale.  The constants here are taken from
 * an approximation of ITU Recommendation 709.  They represent the degree of
 * sensors in our eyes for each color, so there is no purely mathematical
 * "correct" way to do this.
 */
static unsigned char rgb_to_gray(unsigned char r, unsigned char g,
				 unsigned char b)
{
    return (unsigned char) ((227 * r + 724 * g + 73 * b) >> 10);
}

/*
 * Returns a color value if possible, or -1 if not.
 */
static int get_color(char *s)
{
    if (*s == '#') {
	/* RGB color */
	int r = 0, g = 0, b = 0;

#define __read_hexa(x) { \
	s++; \
	if (*s >= 'a') \
	    x = (16 * x) + (*s - 'a' + 10); \
	else if (*s >= 'A') \
	    x = (16 * x) + (*s - 'A' + 10); \
	else if (*s >= '0') \
	    x = (16 * x) + (*s - '0'); \
}
	__read_hexa(r);
	__read_hexa(r);

	__read_hexa(g);
	__read_hexa(g);

	__read_hexa(b);
	__read_hexa(b);

	/*
	 * Invert the color, since 0xFF is black on the HP calculator screen
	 * and this is backwards from common image formats.
	 */
	return 0xFF - rgb_to_gray(r, g, b);
    } else {
	/* No known form of color (HSV is currently unimplemented). */
	return -1;
    }
}

struct hpg_xpm_info _hpg_get_xpm_info(char *xpm[])
{
    struct hpg_xpm_info info;

    /*
     * Read size and format information from the header.
     */
    char *s = xpm[0];

    while (isspace(*s))
	s++;

    info.width = 0;
    while ((!isspace(*s)) && (*s != '\0')) {
	int digit = *(s++) - '0';
	info.width = (10 * info.width) + digit;
    }

    while (isspace(*s))
	s++;

    info.height = 0;
    while ((!isspace(*s)) && (*s != '\0')) {
	int digit = *(s++) - '0';
	info.height = (10 * info.height) + digit;
    }

    while (isspace(*s))
	s++;

    info.ncolors = 0;
    while ((!isspace(*s)) && (*s != '\0')) {
	int digit = *(s++) - '0';
	info.ncolors = (10 * info.ncolors) + digit;
    }

    while (isspace(*s))
	s++;

    info.cpp = 0;
    while ((!isspace(*s)) && (*s != '\0')) {
	int digit = *(s++) - '0';
	info.cpp = (10 * info.cpp) + digit;
    }

    return info;
}

void _hpg_load_xpm(struct hpg_xpm_info info, hpg_t * img, char *xpm[])
{
    struct color {
	char *code;
	unsigned char color;
    };

    struct color *ctable = (struct color *)
	malloc(info.ncolors * sizeof(struct color));
    if (!ctable) {
	perror(__FUNCTION__);
	exit(1);
    }
    ctable[0].code =
	(char *) malloc(info.ncolors * (info.cpp + 1) * sizeof(char));
    if (!ctable[0].code) {
	perror(__FUNCTION__);
	exit(1);
    }

    int i;
    for (i = 0; i < info.ncolors; i++) {
	char *s = xpm[1 + i];
	ctable[i].code = ctable[0].code + i * (info.cpp + 1);
	strncpy(ctable[i].code, s, info.cpp);
	ctable[i].code[info.cpp] = '\0';

	s += info.cpp;
	while (isspace(*s))
	    s++;

	unsigned char color = 0;
	char colorlevel = 0;

	while (*s != '\0') {
	    /* Read a color key */
	    char ckey[3];
	    int j = 0;
	    while ((!isspace(*s)) && (*s != '\0') && j < 3)
		ckey[j++] = *(s++);
	    ckey[j] = '\0';

	    char newlevel = -1;
	    if (strcmp(ckey, "c") == 0)
		newlevel = 1;
	    else if (strcmp(ckey, "g") == 0)
		newlevel = 2;
	    else if (strcmp(ckey, "g4") == 0)
		newlevel = 3;

	    if (newlevel > colorlevel) {
		/* Read a color */
		while (isspace(*s))
		    s++;
		int newcolor = get_color(s);
		if (newcolor != -1) {
		    color = (unsigned char) newcolor;
		    colorlevel = newlevel;
		}
	    }

	    /* Skip the color name */
	    while ((!isspace(*s)) && (*s != '\0'))
		s++;

	    while (isspace(*s))
		s++;
	}

	ctable[i].color = color;
    }

    /*  i is now an index to the current line of the pixmap source */
    i++;
    int y;
    for (y = 0; y < info.height; y++, i++) {
	int x;
	for (x = 0; x < info.width; x++) {
	    int c = 0;
	    while (c < info.ncolors) {
		if (strncmp(xpm[i] + (x * info.cpp),
			    ctable[c].code, info.cpp) == 0) {
		    hpg_set_color(img, ctable[c].color);
		    hpg_draw_pixel_on(img, x, y);
		    break;
		}
		c++;
	    }
	}
    }

    free(ctable[0].code);
    free(ctable);
    return;
}
