/*
 * Ken Clarkson wrote this.  Copyright (c) 1996 by AT&T..
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 */

/*
 * two-dimensional convex hull
 * read points from filename on stack,
 *      one point per line, as two numbers separated by whitespace
 * on stdout, points on convex hull in order around hull, given
 *      by their numbers in input order
 * the results should be "robust", and not return a wildly wrong hull,
 *	    despite using floating point
 * works in O(n log n); I think a bit faster than Graham scan;
 * 	    somewhat like Procedure 8.2 in Edelsbrunner's "Algorithms in Combinatorial
 *	    Geometry", and very close to:
 *	    A.M. Andrew, "Another Efficient Algorithm for Convex Hulls in Two Dimensions",
 *		Info. Proc. Letters 9, 216-219 (1979)
 *	(See also http://geometryalgorithms.com/Archive/algorithm_0110/algorithm_0110.htm)
 */

#include <hpgcc49.h>
#include <hpgraphics.h>

#define N 10000

typedef float coord;
coord points[N][2], *P[N + 1];	/* an extra position is used */
float bb_t, bb_b, bb_l, bb_r;	//bounding box

// read from file, return number of points or 0 (error)
int read_points(char *filename)
{
	char buf[40];
	FILE *fp;
	int k = 0;

	if((fp = fopen(filename, "r")) == NULL)
		return (0);

	while(fgets(buf,sizeof(buf),fp) != NULL && k < N) {
		double x, y;

		if(sscanf(buf,"%lg %lg", &x, &y) != 2)
			continue;

		points[k][0] = x;
		points[k][1] = y;
		P[k] = points[k];

		if(k == 0) {
			bb_r = points[k][0];
			bb_l = points[k][0];
			bb_t = points[k][1];
			bb_b = points[k][1];
		}

		if(points[k][0] > bb_r) bb_r = points[k][0];
		if(points[k][0] < bb_l) bb_l = points[k][0];
		if(points[k][1] > bb_t) bb_t = points[k][1];
		if(points[k][1] < bb_b) bb_b = points[k][1];

		k++;
	}

	fclose(fp);
	return(k);
}

int *hull_points(coord ** P, int m)
{
	int i;
	int *vp = malloc(m * sizeof(int));

	for(i = 0; i < m; i++) {
		int j = (P[i] - points[0]) / 2;
		vp[i] = j;
	}

	return (vp);
}

int ccw(coord ** P, int i, int j, int k)
{
	coord
		a = P[i][0] - P[j][0], b = P[i][1] - P[j][1],
		c = P[k][0] - P[j][0], d = P[k][1] - P[j][1];
	return a * d - b * c <= 0;	/* true if points i, j, k counterclockwise */
}

#define CMPM(c,A,B) \
	v = (*(coord**)A)[c] - (*(coord**)B)[c];\
	if (v>0) return 1;\
	if (v<0) return -1;

int cmpl(const void *a, const void *b)
{
	float v;
	CMPM(0, a, b);
	CMPM(1, b, a);
	return 0;
}

int cmph(const void *a, const void *b)
{
	return cmpl(b, a);
}

int make_chain(coord ** V, int n, int (*cmp) (const void *, const void *))
{
	int i, j, s = 1;
	coord *t;

	qsort(V, n, sizeof(coord *), cmp);
	for(i = 2; i < n; i++) {
		for(j = s; j >= 1 && ccw(V, i, j, j - 1); j--)
			;
		s = j + 1;
		t = V[s];
		V[s] = V[i];
		V[i] = t;
	}
	return s;
}

int ch2d(coord ** P, int n)
{
	int u = make_chain(P, n, cmpl);	/* make lower hull */
	if(!n)
		return 0;
	P[n] = P[0];
	return u + make_chain(P + u, n - u + 1, cmph);	/* make upper hull */
}

int plotit(int n, int m, int v[], int art)
{
	int i;
	int x, y, xl, yl, xf, yf;
	float w = bb_r - bb_l;
	float h = bb_t - bb_b;
	float scale;

	xl = yl = xf = yf = 0;

	hpg_clear();

	if(w / h > 131.0 / 80.0)
		scale = 130.0 / w;
	else
		scale = 79.0 / h;

	for(i = 0; i < n; i++) {
		if(art == 0) {
			x = (points[i][0] - bb_l) * scale;
			x += (130 - w * scale) / 2;
			y = (points[i][1] - bb_b) * scale;
			y += (79 - h * scale) / 2;
		} else {
			x = ((points[i][0] - bb_l) / w) * 130;
			y = ((points[i][1] - bb_b) / h) * 79;
		}
		y = 79 - y;

		hpg_draw_pixel(x, y);
	}

	for(i = 0; i < m; i++) {
		if(art == 0) {
			x = (points[v[i]][0] - bb_l) * scale;
			x += (130 - w * scale) / 2;
			y = (points[v[i]][1] - bb_b) * scale;
			y += (79 - h * scale) / 2;
		} else {
			x = ((points[v[i]][0] - bb_l) / w) * 130;
			y = ((points[v[i]][1] - bb_b) / h) * 79;
		}
		y = 79 - y;

		if (i == 0) {
			xf = xl = x;
			yf = yl = y;
			continue;
		}

		hpg_draw_line(xl, yl, x, y);

		xl = x;
		yl = y;
	}
	hpg_draw_line(xl, yl, xf, yf);

	return (0);
}

int snapshot(char *grob)
{
	char hex[1];
	int g, t, x, y;

	sprintf(grob, "GROB 131 80 ");

	for(y = 0; y < 80; y++) {
		g = 1;
		t = 0;
		for(x = 0; x < 131; x++) {
			char p;

			if((p = hpg_get_pixel(hpg_stdscreen, x, y)))
				t += g;

			g *= 2;

			if(g == 16) {
				sprintf(hex, "%X", t);
				strcat(grob, hex);
				g = 1;
				t = 0;
			}
		}
		sprintf(hex, "%X", t);
		strcat(grob, hex);
		sprintf(hex, "%X", 0);
		strcat(grob, hex);
	}

	return (0);
}

int main(int argc, char *argv[])
{
	int n, m, i;
#ifndef HPAPINE
	char *hullpoints, grob[2740];
	SAT_LIST p;
#endif
	char *filename;
	double A = 0, CHx = 0, CHy = 0, PM = 0, CPx, CPy;
	int *v, art;
	SAT_STACK_ELEMENT e;
	SAT_STACK_DATA d;

	art = sat_pop_real();
	if(art != 0 && art != 1) {
		sat_push_zint_llong(art);
		sat_stack_push_string("Aspect ratio type must be 0 or 1");
		sat_push_real(1);
		return(0);
	}

	sat_get_stack_element(1, &e);
	sat_decode_stack_element(&d, &e);

	if(d.type == SAT_DATA_TYPE_STRING) {
		filename = str_unquote(d.sval, '\'');
		sat_stack_drop();
	} else {
		sat_stack_push_string("Missing Filename");
		sat_push_real(1);
		return (0);
	}

	sys_slowOff();

	hpg_clear();
	hpg_set_font(hpg_stdscreen, hpg_get_bigfont());
	hpg_draw_text("Computing Hull...", 15, 38);
	hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);

	n = read_points(filename);
	if(n == 0) {
		char errormsg[30];

		sprintf(errormsg,"Cannot open: %s",filename);
		sat_stack_push_string(errormsg);
		sat_push_real(1);
		sys_slowOn();
		return (0);
	}
	if(n < 3) {
		sat_stack_push_string("Number of points must be > 2!");
		sat_push_real(1);
		sys_slowOn();
		return (0);
	}

	m = ch2d(P, n);
	v = hull_points(P, m);

	for(i = 0; i < m; i++) {
		double C, Px, Py;
		int j = (i + 1) % m;
		A += points[v[i]][0] * points[v[j]][1];
		A -= points[v[i]][1] * points[v[j]][0];
		C = ((points[v[i]][0] * points[v[j]][1]) -
			 (points[v[j]][0] * points[v[i]][1]));
		CHx += (points[v[i]][0] + points[v[j]][0]) * C;
		CHy += (points[v[i]][1] + points[v[j]][1]) * C;
		Px = (points[v[j]][0] - points[v[i]][0]);
		Py = (points[v[j]][1] - points[v[i]][1]);
		PM += sqrt(Px * Px + Py * Py);
	}
	A /= 2;
	CHx /= (6 * A);
	CHy /= (6 * A);

	CPx = points[0][0];
	CPy = points[0][1];
	for(i = 1; i < n; i++) {
		CPx = (CPx * i + points[i][0]) / (i + 1);
		CPy = (CPy * i + points[i][1]) / (i + 1);
	}

	plotit(n, m, v, art);
#ifndef HPAPINE
	snapshot(grob);

	if((hullpoints = malloc((38 * m + 4) * sizeof(char))) == NULL) {
		sat_stack_push_string("Failed to allocate memory for hull points.");
		sat_push_real(1);
		return (0);
	}
	hullpoints[0] = '\0';
	strcat(hullpoints, "{ ");
	for(i = 0; i < m; i++) {
		char point[38];
		sprintf(point, "{ %0.6E %0.6E } ", points[v[i]][0],
				points[v[i]][1]);
		strcat(hullpoints, point);
	}
	strcat(hullpoints, " }");

	sat_stack_push_string(grob);
	sat_stack_push_string(hullpoints);
	sat_push_zint_llong(n);
	sat_stack_push_string("n");
	sat_push_real(A);
	sat_stack_push_string("Area");
	sat_push_real(PM);
	sat_stack_push_string("Perimeter");
	p = sat_list_create(20);
	sat_list_add_real(p, CPx);
	sat_list_add_real(p, CPy);
	sat_push_list(p);
	sat_list_destroy(p);
	sat_stack_push_string("COG(p)");
	p = sat_list_create(20);
	sat_list_add_real(p, CHx);
	sat_list_add_real(p, CHy);
	sat_push_list(p);
	sat_list_destroy(p);
	sat_stack_push_string("COG(h)");
	sat_push_real(0);
	hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_WHITE);
	beep();
	sys_waitRTCTicks(1);
#endif
	sys_slowOn();
	WAIT_CANCEL;

	return (0);
}
