1     	/*
2     	 * Ken Clarkson wrote this.  Copyright (c) 1996 by AT&T..
3     	 * Permission to use, copy, modify, and distribute this software for any
4     	 * purpose without fee is hereby granted, provided that this entire notice
5     	 * is included in all copies of any software which is or includes a copy
6     	 * or modification of this software and in all copies of the supporting
7     	 * documentation for such software.
8     	 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
9     	 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
10    	 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
11    	 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
12    	 */
13    	
14    	/*
15    	 * two-dimensional convex hull
16    	 * read points from filename on stack,
17    	 *      one point per line, as two numbers separated by whitespace
18    	 * on stdout, points on convex hull in order around hull, given
19    	 *      by their numbers in input order
20    	 * the results should be "robust", and not return a wildly wrong hull,
21    	 *        despite using floating point
22    	 * works in O(n log n); I think a bit faster than Graham scan;
23    	 *         somewhat like Procedure 8.2 in Edelsbrunner's "Algorithms in Combinatorial
24    	 *        Geometry", and very close to:
25    	 *        A.M. Andrew, "Another Efficient Algorithm for Convex Hulls in Two Dimensions",
26    	 *        Info. Proc. Letters 9, 216-219 (1979)
27    	 *    (See also http://geometryalgorithms.com/Archive/algorithm_0110/algorithm_0110.htm)
28    	 */
29    	
30    	#include <hpgcc49.h>
31    	#include <hpgraphics.h>
32    	
33    	#define N 10000
34    	
35    	typedef float coord;
36    	coord points[N][2], *P[N + 1];    /* an extra position is used */
37    	float bb_t, bb_b, bb_l, bb_r;    //bounding box
38    	
39    	// read from file, return number of points or 0 (error)
40    	int read_points(char *filename)
41    	{
42    	    char buf[40];
43    	    FILE *fp;
44    	    int k = 0;
45    	
46    	    if((fp = fopen(filename, "r")) == NULL)
47    	        return (0);
48    	
49    	    while(fgets(buf,sizeof(buf),fp) != NULL && k < N) {
50    	        double x, y;
51    	
52    	        if(sscanf(buf,"%lg %lg", &x, &y) != 2)
53    	            continue;
54    	
55    	        points[k][0] = x;
56    	        points[k][1] = y;
57    	        P[k] = points[k];
58    	
59    	        if(k == 0) {
60    	            bb_r = points[k][0];
61    	            bb_l = points[k][0];
62    	            bb_t = points[k][1];
63    	            bb_b = points[k][1];
64    	        }
65    	
66    	        if(points[k][0] > bb_r) bb_r = points[k][0];
67    	        if(points[k][0] < bb_l) bb_l = points[k][0];
68    	        if(points[k][1] > bb_t) bb_t = points[k][1];
69    	        if(points[k][1] < bb_b) bb_b = points[k][1];
70    	
71    	        k++;
72    	    }
73    	
74    	    fclose(fp);
75    	    return(k);
76    	}
77    	
78    	int *hull_points(coord ** P, int m)
79    	{
80    	    int i;
81    	    int *vp = malloc(m * sizeof(int));
82    	
83    	    for(i = 0; i < m; i++) {
84    	        int j = (P[i] - points[0]) / 2;
85    	        vp[i] = j;
86    	    }
87    	
88    	    return (vp);
89    	}
90    	
91    	int ccw(coord ** P, int i, int j, int k)
92    	{
93    	    coord
94    	        a = P[i][0] - P[j][0], b = P[i][1] - P[j][1],
95    	        c = P[k][0] - P[j][0], d = P[k][1] - P[j][1];
96    	    return a * d - b * c <= 0;    /* true if points i, j, k counterclockwise */
97    	}
98    	
99    	#define CMPM(c,A,B) \
100   	    v = (*(coord**)A)[c] - (*(coord**)B)[c];\
101   	    if (v>0) return 1;\
102   	    if (v<0) return -1;
103   	
104   	int cmpl(const void *a, const void *b)
105   	{
106   	    float v;
107   	    CMPM(0, a, b);
108   	    CMPM(1, b, a);
109   	    return 0;
110   	}
111   	
112   	int cmph(const void *a, const void *b)
113   	{
114   	    return cmpl(b, a);
115   	}
116   	
117   	int make_chain(coord ** V, int n, int (*cmp) (const void *, const void *))
118   	{
119   	    int i, j, s = 1;
120   	    coord *t;
121   	
122   	    qsort(V, n, sizeof(coord *), cmp);
123   	    for(i = 2; i < n; i++) {
124   	        for(j = s; j >= 1 && ccw(V, i, j, j - 1); j--)
125   	            ;
126   	        s = j + 1;
127   	        t = V[s];
128   	        V[s] = V[i];
129   	        V[i] = t;
130   	    }
131   	    return s;
132   	}
133   	
134   	int ch2d(coord ** P, int n)
135   	{
136   	    int u = make_chain(P, n, cmpl);    /* make lower hull */
137   	    if(!n)
138   	        return 0;
139   	    P[n] = P[0];
140   	    return u + make_chain(P + u, n - u + 1, cmph);    /* make upper hull */
141   	}
142   	
143   	int plotit(int n, int m, int v[], int art)
144   	{
145   	    int i;
146   	    int x, y, xl, yl, xf, yf;
147   	    float w = bb_r - bb_l;
148   	    float h = bb_t - bb_b;
149   	    float scale;
150   	
151   	    xl = yl = xf = yf = 0;
152   	
153   	    hpg_clear();
154   	
155   	    if(w / h > 131.0 / 80.0)
156   	        scale = 130.0 / w;
157   	    else
158   	        scale = 79.0 / h;
159   	
160   	    for(i = 0; i < n; i++) {
161   	        if(art == 0) {
162   	            x = (points[i][0] - bb_l) * scale;
163   	            x += (130 - w * scale) / 2;
164   	            y = (points[i][1] - bb_b) * scale;
165   	            y += (79 - h * scale) / 2;
166   	        } else {
167   	            x = ((points[i][0] - bb_l) / w) * 130;
168   	            y = ((points[i][1] - bb_b) / h) * 79;
169   	        }
170   	        y = 79 - y;
171   	
172   	        hpg_draw_pixel(x, y);
173   	    }
174   	
175   	    for(i = 0; i < m; i++) {
176   	        if(art == 0) {
177   	            x = (points[v[i]][0] - bb_l) * scale;
178   	            x += (130 - w * scale) / 2;
179   	            y = (points[v[i]][1] - bb_b) * scale;
180   	            y += (79 - h * scale) / 2;
181   	        } else {
182   	            x = ((points[v[i]][0] - bb_l) / w) * 130;
183   	            y = ((points[v[i]][1] - bb_b) / h) * 79;
184   	        }
185   	        y = 79 - y;
186   	
187   	        if (i == 0) {
188   	            xf = xl = x;
189   	            yf = yl = y;
190   	            continue;
191   	        }
192   	
193   	        hpg_draw_line(xl, yl, x, y);
194   	
195   	        xl = x;
196   	        yl = y;
197   	    }
198   	    hpg_draw_line(xl, yl, xf, yf);
199   	
200   	    return (0);
201   	}
202   	
203   	int snapshot(char *grob)
204   	{
205   	    char hex[1];
206   	    int g, t, x, y;
207   	
208   	    sprintf(grob, "GROB 131 80 ");
209   	
210   	    for(y = 0; y < 80; y++) {
211   	        g = 1;
212   	        t = 0;
213   	        for(x = 0; x < 131; x++) {
214   	            char p;
215   	
216   	            if((p = hpg_get_pixel(hpg_stdscreen, x, y)))
217   	                t += g;
218   	
219   	            g *= 2;
220   	
221   	            if(g == 16) {
222   	                sprintf(hex, "%X", t);
223   	                strcat(grob, hex);
224   	                g = 1;
225   	                t = 0;
226   	            }
227   	        }
228   	        sprintf(hex, "%X", t);
229   	        strcat(grob, hex);
230   	        sprintf(hex, "%X", 0);
231   	        strcat(grob, hex);
232   	    }
233   	
234   	    return (0);
235   	}
236   	
237   	int main(int argc, char *argv[])
238   	{
239   	    int n, m, i;
240   	#ifndef HPAPINE
241   	    char *hullpoints, grob[2740];
242   	    SAT_LIST p;
243   	#endif
244   	    char *filename;
245   	    double A = 0, CHx = 0, CHy = 0, PM = 0, CPx, CPy;
246   	    int *v, art;
247   	    SAT_STACK_ELEMENT e;
248   	    SAT_STACK_DATA d;
249   	
250   	    art = sat_pop_real();
251   	    if(art != 0 && art != 1) {
252   	        sat_push_zint_llong(art);
253   	        sat_stack_push_string("Aspect ratio type must be 0 or 1");
254   	        sat_push_real(1);
255   	        return(0);
256   	    }
257   	
258   	    sat_get_stack_element(1, &e);
259   	    sat_decode_stack_element(&d, &e);
260   	
261   	    if(d.type == SAT_DATA_TYPE_STRING) {
262   	        filename = str_unquote(d.sval, '\'');
263   	        sat_stack_drop();
264   	    } else {
265   	        sat_stack_push_string("Missing Filename");
266   	        sat_push_real(1);
267   	        return (0);
268   	    }
269   	
270   	    sys_slowOff();
271   	
272   	    hpg_clear();
273   	    hpg_set_font(hpg_stdscreen, hpg_get_bigfont());
274   	    hpg_draw_text("Computing Hull...", 15, 38);
275   	    hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);
276   	
277   	    n = read_points(filename);
278   	    if(n == 0) {
279   	        char errormsg[30];
280   	
281   	        sprintf(errormsg,"Cannot open: %s",filename);
282   	        sat_stack_push_string(errormsg);
283   	        sat_push_real(1);
284   	        sys_slowOn();
285   	        return (0);
286   	    }
287   	    if(n < 3) {
288   	        sat_stack_push_string("Number of points must be > 2!");
289   	        sat_push_real(1);
290   	        sys_slowOn();
291   	        return (0);
292   	    }
293   	
294   	    m = ch2d(P, n);
295   	    v = hull_points(P, m);
296   	
297   	    for(i = 0; i < m; i++) {
298   	        double C, Px, Py;
299   	        int j = (i + 1) % m;
300   	        A += points[v[i]][0] * points[v[j]][1];
301   	        A -= points[v[i]][1] * points[v[j]][0];
302   	        C = ((points[v[i]][0] * points[v[j]][1]) -
303   	             (points[v[j]][0] * points[v[i]][1]));
304   	        CHx += (points[v[i]][0] + points[v[j]][0]) * C;
305   	        CHy += (points[v[i]][1] + points[v[j]][1]) * C;
306   	        Px = (points[v[j]][0] - points[v[i]][0]);
307   	        Py = (points[v[j]][1] - points[v[i]][1]);
308   	        PM += sqrt(Px * Px + Py * Py);
309   	    }
310   	    A /= 2;
311   	    CHx /= (6 * A);
312   	    CHy /= (6 * A);
313   	
314   	    CPx = points[0][0];
315   	    CPy = points[0][1];
316   	    for(i = 1; i < n; i++) {
317   	        CPx = (CPx * i + points[i][0]) / (i + 1);
318   	        CPy = (CPy * i + points[i][1]) / (i + 1);
319   	    }
320   	
321   	    plotit(n, m, v, art);
322   	#ifndef HPAPINE
323   	    snapshot(grob);
324   	
325   	    if((hullpoints = malloc((50 * m + 4) * sizeof(char))) == NULL) {
326   	        sat_stack_push_string("Failed to allocate memory for hull points.");
327   	        sat_push_real(1);
328   	        return (0);
329   	    }
330   	    hullpoints[0] = '\0';
331   	    strcat(hullpoints, "{ ");
332   	    for(i = 0; i < m; i++) {
333   	        char point[50];
334   	        sprintf(point, "{ %0.12E %0.12E } ", points[v[i]][0],
335   	                points[v[i]][1]);
336   	        strcat(hullpoints, point);
337   	    }
338   	    strcat(hullpoints, " }");
339   	
340   	    sat_stack_push_string(grob);
341   	    sat_stack_push_string(hullpoints);
342   	    sat_push_zint_llong(n);
343   	    sat_stack_push_string("n");
344   	    sat_push_real(A);
345   	    sat_stack_push_string("Area");
346   	    sat_push_real(PM);
347   	    sat_stack_push_string("Perimeter");
348   	    p = sat_list_create(20);
349   	    sat_list_add_real(p, CPx);
350   	    sat_list_add_real(p, CPy);
351   	    sat_push_list(p);
352   	    sat_list_destroy(p);
353   	    sat_stack_push_string("COG(p)");
354   	    p = sat_list_create(20);
355   	    sat_list_add_real(p, CHx);
356   	    sat_list_add_real(p, CHy);
357   	    sat_push_list(p);
358   	    sat_list_destroy(p);
359   	    sat_stack_push_string("COG(h)");
360   	    sat_push_real(0);
361   	    hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_WHITE);
362   	    beep();
363   	    sys_waitRTCTicks(1);
364   	#endif
365   	    sys_slowOn();
366   	    WAIT_CANCEL;
367   	
368   	    return (0);
369   	}
