/*
 * Ken Clarkson wrote this.  Copyright (c) 1996 by AT&T..
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 */

/*
 * two-dimensional convex hull
 * read points from filename on stack,
 *      one point per line, as two numbers separated by whitespace
 * on stdout, points on convex hull in order around hull, given
 *      by their numbers in input order
 * the results should be "robust", and not return a wildly wrong hull,
 *	    despite using floating point
 * works in O(n log n); I think a bit faster than Graham scan;
 * 	    somewhat like Procedure 8.2 in Edelsbrunner's "Algorithms in Combinatorial
 *	    Geometry", and very close to:
 *	    A.M. Andrew, "Another Efficient Algorithm for Convex Hulls in Two Dimensions",
 *		Info. Proc. Letters 9, 216-219 (1979)
 *	(See also http://geometryalgorithms.com/Archive/algorithm_0110/algorithm_0110.htm)
 */

#include <hpgcc49.h>
#include <hpgraphics.h>
#include <hpgbmp.h>

#define N 10000
int W;
int H;

typedef float coord;
coord points[N][2], *P[N + 1];	/* an extra position is used */
float bb_t, bb_b, bb_l, bb_r;	//bounding box

// read from file, return number of points or 0 (error)
int read_points(char *filename)
{
	char buf[40];
	FILE *fp;
	int k = 0;

	if((fp = fopen(filename, "r")) == NULL)
		return (0);

	while(fgets(buf,sizeof(buf),fp) != NULL && k < N) {
		double x, y;

		if(sscanf(buf,"%lg %lg", &x, &y) != 2)
			continue;

		points[k][0] = x;
		points[k][1] = y;
		P[k] = points[k];

		if(k == 0) {
			bb_r = points[k][0];
			bb_l = points[k][0];
			bb_t = points[k][1];
			bb_b = points[k][1];
		}

		if(points[k][0] > bb_r) bb_r = points[k][0];
		if(points[k][0] < bb_l) bb_l = points[k][0];
		if(points[k][1] > bb_t) bb_t = points[k][1];
		if(points[k][1] < bb_b) bb_b = points[k][1];

		k++;
	}

	fclose(fp);
	return(k);
}

int *hull_points(coord ** P, int m)
{
	int i;
	int *vp = malloc(m * sizeof(int));

	for(i = 0; i < m; i++) {
		int j = (P[i] - points[0]) / 2;
		vp[i] = j;
	}

	return (vp);
}

int ccw(coord ** P, int i, int j, int k)
{
	coord
		a = P[i][0] - P[j][0], b = P[i][1] - P[j][1],
		c = P[k][0] - P[j][0], d = P[k][1] - P[j][1];
	return a * d - b * c <= 0;	/* true if points i, j, k counterclockwise */
}

#define CMPM(c,A,B) \
	v = (*(coord**)A)[c] - (*(coord**)B)[c];\
	if (v>0) return 1;\
	if (v<0) return -1;

int cmpl(const void *a, const void *b)
{
	float v;
	CMPM(0, a, b);
	CMPM(1, b, a);
	return 0;
}

int cmph(const void *a, const void *b)
{
	return cmpl(b, a);
}

int make_chain(coord ** V, int n, int (*cmp) (const void *, const void *))
{
	int i, j, s = 1;
	coord *t;

	qsort(V, n, sizeof(coord *), cmp);
	for(i = 2; i < n; i++) {
		for(j = s; j >= 1 && ccw(V, i, j, j - 1); j--)
			;
		s = j + 1;
		t = V[s];
		V[s] = V[i];
		V[i] = t;
	}
	return s;
}

int ch2d(coord ** P, int n)
{
	int u = make_chain(P, n, cmpl);	/* make lower hull */
	if(!n)
		return 0;
	P[n] = P[0];
	return u + make_chain(P + u, n - u + 1, cmph);	/* make upper hull */
}

int plotit(int n, int m, int v[], int art, hpg_t *image, double CHx, double CHy, double CPx, double CPy)
{
	int i;
	int x, y, xl, yl, xf, yf;
	float w = bb_r - bb_l;
	float h = bb_t - bb_b;
	float scale;

	xl = yl = xf = yf = 0;

	hpg_clear_on(image);
	hpg_set_color(image,HPG_COLOR_GRAY_13);

	if(w / h > (float)W / (float)H)
		scale = (W-1) / w;
	else
		scale = (H-1) / h;

	for(i = 0; i < n; i++) {
		if(art == 0) {
			x = (points[i][0] - bb_l) * scale;
			x += ((W-1) - w * scale) / 2;
			y = (points[i][1] - bb_b) * scale;
			y += ((H-1) - h * scale) / 2;
		} else {
			x = ((points[i][0] - bb_l) / w) * (W-1);
			y = ((points[i][1] - bb_b) / h) * (H-1);
		}
		y = (H-1) - y;

		hpg_draw_line_on(image,x+1, y, x-1, y);
		hpg_draw_line_on(image,x, y+1, x, y-1);
	}

	for(i = 0; i < m; i++) {
		if(art == 0) {
			x = (points[v[i]][0] - bb_l) * scale;
			x += ((W-1) - w * scale) / 2;
			y = (points[v[i]][1] - bb_b) * scale;
			y += ((H-1) - h * scale) / 2;
		} else {
			x = ((points[v[i]][0] - bb_l) / w) * (W-1);
			y = ((points[v[i]][1] - bb_b) / h) * (H-1);
		}
		y = (H-1) - y;

		if (i == 0) {
			xf = xl = x;
			yf = yl = y;
			continue;
		}

		hpg_set_color(image,HPG_COLOR_GRAY_14);
		hpg_draw_line_on(image,xl, yl, x, y);

		xl = x;
		yl = y;
	}
	hpg_draw_line_on(image,xl, yl, xf, yf);

	hpg_set_color(image,HPG_COLOR_BLACK);
	if(art == 0) {
		x = (CHx - bb_l) * scale;
		x += ((W-1) - w * scale) / 2;
		y = (CHy - bb_b) * scale;
		y += ((H-1) - h * scale) / 2;
	} else {
		x = ((CHx - bb_l) / w) * (W-1);
		y = ((CHy - bb_b) / h) * (H-1);
	}
	y = (H-1) - y;
	hpg_fill_circle_on(image,x,y,3);

	hpg_set_color(image,HPG_COLOR_GRAY_12);
	if(art == 0) {
		x = (CPx - bb_l) * scale;
		x += ((W-1) - w * scale) / 2;
		y = (CPy - bb_b) * scale;
		y += ((H-1) - h * scale) / 2;
	} else {
		x = ((CPx - bb_l) / w) * (W-1);
		y = ((CPy - bb_b) / h) * (H-1);
	}
	y = (H-1) - y;
	hpg_fill_rect_on(image,x-2,y-2,x+2,y+2);

	return (0);
}

int grobshot(char *filename, hpg_t *image)
{
	int g, t, x, y;
	unsigned char header[10], p;
	unsigned long int length;
	FILE *fp;

	if((fp = fopen(filename,"wb")) == NULL) {
		return(1);
	}

	fwrite("HPHP49-C",1,8,fp);

	header[0] = 0x1e;
	header[1] = 0x2b;

	length = (int)ceil(W/8.0) * 2 * H + 15;

	header[2] = (length & 0x0f) << 4;
	header[3] = (length >> 4) & 0xff;
	header[4] = (length >> 12) & 0xff;

	header[5] = H & 0xff;
	header[6] = (H >> 8) & 0xff;
	header[7] = ((W & 0x0f) << 4) | ((H >> 16) & 0x0f);
	header[8] = (W >> 4) & 0xff;
	header[9] = (W >> 12) & 0xff;

	fwrite(header,1,10,fp);

	for(y = 0; y < H; y++) {
		g = 1;
		t = 0;
		for(x = 0; x < W; x++) {
			if((p = hpg_get_pixel(image, x, y)))
				t += g;

			g *= 2;

			if(g == 256) {
				fwrite(&t,1,1,fp);
				t = 0;
				g = 1;
			}
		}
		if(g != 1)
			fwrite(&t,1,1,fp);
	}

	fclose(fp);
	return (0);
}

int main(int argc, char *argv[])
{
	int n, m, i, j;
	char *hullpoints, *filename, *ofilename, *ofiletype, msg[40];
	double A = 0, CHx = 0, CHy = 0, PM = 0, CPx, CPy;
	int *v, art, grob=0, bmp=0;
#ifndef HPAPINE
	SAT_LIST p;
#endif
	SAT_STACK_ELEMENT e;
	SAT_STACK_DATA d;
	hpg_t *image;
	FILE *fp;

	art = sat_pop_real();
	if(art != 0 && art != 1) {
		sat_stack_push_string("Aspect ratio type must be 0 or 1");
		sat_push_real(1);
		return(0);
	}

	H = sat_pop_real();
	if(H < 1) {
		sat_stack_push_string("Height > 0");
		sat_push_real(1);
		return(0);
	}

	W = sat_pop_real();
	if(W < 1) {
		sat_stack_push_string("Width > 0");
		sat_push_real(1);
		return(0);
	}

	sat_get_stack_element(1, &e);
	sat_decode_stack_element(&d, &e);

	if(d.type == SAT_DATA_TYPE_STRING) {
		ofilename = str_unquote(d.sval, '\'');
		sat_stack_drop();
	} else {
		sat_stack_push_string("Missing output filename");
		sat_push_real(1);
		return (0);
	}

	if(strlen(ofilename) > 4)
		ofiletype = ofilename + strlen(ofilename) - 3;

	if(strcmp(ofiletype,"GRO") == 0)
		grob = 1;
	if(strcmp(ofiletype,"BMP") == 0)
		bmp = 1;

	if(grob + bmp != 1) {
		sat_stack_push_string("File ext. must be .GRO or .BMP");
		sat_push_real(1);
		return (0);
	}

	sat_get_stack_element(1, &e);
	sat_decode_stack_element(&d, &e);

	if(d.type == SAT_DATA_TYPE_STRING) {
		filename = str_unquote(d.sval, '\'');
		sat_stack_drop();
	} else {
		sat_stack_push_string("Missing input filename");
		sat_push_real(1);
		return (0);
	}

	sys_slowOff();

	if((image = hpg_alloc_gray16_image(W,H)) == NULL) {
		sprintf(msg,"Failed to allocate %dx%d image",W,H);
		sat_stack_push_string(msg);
		sat_push_real(1);
		return(0);
	}

	if((fp = fopen(ofilename,"w")) == NULL) {
		sprintf(msg,"Cannot open: %s",ofilename);
		sat_stack_push_string(msg);
		sat_push_real(1);
		return(0);
	}
	fclose(fp);

	hpg_clear();
	hpg_set_font(hpg_stdscreen, hpg_get_bigfont());
	hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);
	hpg_draw_text("Computing Hull...", 15, 38);

	n = read_points(filename);
	if(n == 0) {
		sprintf(msg,"Cannot read: %s",filename);
		sat_stack_push_string(msg);
		sat_push_real(1);
		sys_slowOn();
		return (0);
	}
	if(n < 3) {
		sat_stack_push_string("Number of points must be > 2!");
		sat_push_real(1);
		sys_slowOn();
		return (0);
	}

	m = ch2d(P, n);
	v = hull_points(P, m);

	for(i = 0; i < m; i++) {
		double C, Px, Py;
		int j = (i + 1) % m;
		A += points[v[i]][0] * points[v[j]][1];
		A -= points[v[i]][1] * points[v[j]][0];
		C = ((points[v[i]][0] * points[v[j]][1]) -
			 (points[v[j]][0] * points[v[i]][1]));
		CHx += (points[v[i]][0] + points[v[j]][0]) * C;
		CHy += (points[v[i]][1] + points[v[j]][1]) * C;
		Px = (points[v[j]][0] - points[v[i]][0]);
		Py = (points[v[j]][1] - points[v[i]][1]);
		PM += sqrt(Px * Px + Py * Py);
	}
	A /= 2;
	CHx /= (6 * A);
	CHy /= (6 * A);

	CPx = points[0][0];
	CPy = points[0][1];
	for(i = 1; i < n; i++) {
		CPx = (CPx * i + points[i][0]) / (i + 1);
		CPy = (CPy * i + points[i][1]) / (i + 1);
	}

	plotit(n, m, v, art, image, CHx, CHy, CPx, CPy);
	hpg_clear();
	hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);

	if(grob) {
		hpg_draw_text("Computing GROB...", 15, 38);
		if(grobshot(ofilename, image) == 1) {
			sprintf(msg,"Cannot open: %s",ofilename);
			sat_stack_push_string(msg);
			sat_push_real(1);
			return(0);
		}
	}

#define R 2
#define G 1
#define B 0

	if(bmp) {
		int ret, i;
		palette_t mycolors[16];

		for(i = 0;i < 16;i++)
			for(j = 0;j < 4;j++)
				mycolors[i].rgb[j] = 0x00;

		//background
		mycolors[0].rgb[B] = 0xff;
		mycolors[0].rgb[G] = 0xff;
		mycolors[0].rgb[R] = 0xff;

		//green COG(p)
		mycolors[12].rgb[R] = 0x00;
		mycolors[12].rgb[G] = 0xff;
		mycolors[12].rgb[B] = 0x00;

		//red sites
		mycolors[13].rgb[R] = 0xff;
		mycolors[13].rgb[G] = 0x00;
		mycolors[13].rgb[B] = 0x00;

		//blue CH
		mycolors[14].rgb[R] = 0x00;
		mycolors[14].rgb[G] = 0x00;
		mycolors[14].rgb[B] = 0xff;

		//black COG(h)
		mycolors[15].rgb[R] = 0x00;
		mycolors[15].rgb[G] = 0x00;
		mycolors[15].rgb[B] = 0x00;

		hpg_draw_text("Computing BMP...", 15, 38);
		if((ret = hpg_save_anybmp_gray16(ofilename,mycolors,image,W,H)) != C_NoError) {
			if(ret == C_ErrOpenFile)
				sprintf(msg,"Cannot open: %s",ofilename);
			if(ret == C_ErrMemory)
				sprintf(msg,"Out of memory");
			if(ret == C_ErrWritePalette)
				sprintf(msg,"Unable to write palette");

			sat_stack_push_string(msg);
			sat_push_real(1);
			return(0);
		}
	}

	if((hullpoints = malloc((50 * m + 4) * sizeof(char))) == NULL) {
		sat_stack_push_string("Failed to allocate memory for hull points.");
		sat_push_real(1);
		return (0);
	}
	hullpoints[0] = '\0';
	strcat(hullpoints, "{ ");
	for(i = 0; i < m; i++) {
		char point[50];
		sprintf(point, "{ %0.12E %0.12E } ", points[v[i]][0],
				points[v[i]][1]);
		strcat(hullpoints, point);
	}
	strcat(hullpoints, " }");

	sat_stack_push_string(hullpoints);
	sat_push_zint_llong(n);
	sat_stack_push_string("n");
	sat_push_real(A);
	sat_stack_push_string("Area");
	sat_push_real(PM);
	sat_stack_push_string("Perimeter");
#ifndef HPAPINE
	p = sat_list_create(20);
	sat_list_add_real(p, CPx);
	sat_list_add_real(p, CPy);
	sat_push_list(p);
	sat_list_destroy(p);
	sat_stack_push_string("COG(p)");
	p = sat_list_create(20);
	sat_list_add_real(p, CHx);
	sat_list_add_real(p, CHy);
	sat_push_list(p);
	sat_list_destroy(p);
	sat_stack_push_string("COG(h)");
#endif
	sat_push_real(0);
	hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_WHITE);
	sys_waitRTCTicks(1);
	sys_slowOn();

	return (0);
}
