1     	/*
2     	 * Ken Clarkson wrote this.  Copyright (c) 1996 by AT&T..
3     	 * Permission to use, copy, modify, and distribute this software for any
4     	 * purpose without fee is hereby granted, provided that this entire notice
5     	 * is included in all copies of any software which is or includes a copy
6     	 * or modification of this software and in all copies of the supporting
7     	 * documentation for such software.
8     	 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
9     	 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
10    	 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
11    	 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
12    	 */
13    	
14    	/*
15    	 * two-dimensional convex hull
16    	 * read points from filename on stack,
17    	 *      one point per line, as two numbers separated by whitespace
18    	 * on stdout, points on convex hull in order around hull, given
19    	 *      by their numbers in input order
20    	 * the results should be "robust", and not return a wildly wrong hull,
21    	 *        despite using floating point
22    	 * works in O(n log n); I think a bit faster than Graham scan;
23    	 *         somewhat like Procedure 8.2 in Edelsbrunner's "Algorithms in Combinatorial
24    	 *        Geometry", and very close to:
25    	 *        A.M. Andrew, "Another Efficient Algorithm for Convex Hulls in Two Dimensions",
26    	 *        Info. Proc. Letters 9, 216-219 (1979)
27    	 *    (See also http://geometryalgorithms.com/Archive/algorithm_0110/algorithm_0110.htm)
28    	 */
29    	
30    	#include <hpgcc49.h>
31    	#include <hpgraphics.h>
32    	#include <hpgbmp.h>
33    	
34    	#define N 10000
35    	int W;
36    	int H;
37    	
38    	typedef float coord;
39    	coord points[N][2], *P[N + 1];    /* an extra position is used */
40    	float bb_t, bb_b, bb_l, bb_r;    //bounding box
41    	
42    	// read from file, return number of points or 0 (error)
43    	int read_points(char *filename)
44    	{
45    	    char buf[40];
46    	    FILE *fp;
47    	    int k = 0;
48    	
49    	    if((fp = fopen(filename, "r")) == NULL)
50    	        return (0);
51    	
52    	    while(fgets(buf,sizeof(buf),fp) != NULL && k < N) {
53    	        double x, y;
54    	
55    	        if(sscanf(buf,"%lg %lg", &x, &y) != 2)
56    	            continue;
57    	
58    	        points[k][0] = x;
59    	        points[k][1] = y;
60    	        P[k] = points[k];
61    	
62    	        if(k == 0) {
63    	            bb_r = points[k][0];
64    	            bb_l = points[k][0];
65    	            bb_t = points[k][1];
66    	            bb_b = points[k][1];
67    	        }
68    	
69    	        if(points[k][0] > bb_r) bb_r = points[k][0];
70    	        if(points[k][0] < bb_l) bb_l = points[k][0];
71    	        if(points[k][1] > bb_t) bb_t = points[k][1];
72    	        if(points[k][1] < bb_b) bb_b = points[k][1];
73    	
74    	        k++;
75    	    }
76    	
77    	    fclose(fp);
78    	    return(k);
79    	}
80    	
81    	int *hull_points(coord ** P, int m)
82    	{
83    	    int i;
84    	    int *vp = malloc(m * sizeof(int));
85    	
86    	    for(i = 0; i < m; i++) {
87    	        int j = (P[i] - points[0]) / 2;
88    	        vp[i] = j;
89    	    }
90    	
91    	    return (vp);
92    	}
93    	
94    	int ccw(coord ** P, int i, int j, int k)
95    	{
96    	    coord
97    	        a = P[i][0] - P[j][0], b = P[i][1] - P[j][1],
98    	        c = P[k][0] - P[j][0], d = P[k][1] - P[j][1];
99    	    return a * d - b * c <= 0;    /* true if points i, j, k counterclockwise */
100   	}
101   	
102   	#define CMPM(c,A,B) \
103   	    v = (*(coord**)A)[c] - (*(coord**)B)[c];\
104   	    if (v>0) return 1;\
105   	    if (v<0) return -1;
106   	
107   	int cmpl(const void *a, const void *b)
108   	{
109   	    float v;
110   	    CMPM(0, a, b);
111   	    CMPM(1, b, a);
112   	    return 0;
113   	}
114   	
115   	int cmph(const void *a, const void *b)
116   	{
117   	    return cmpl(b, a);
118   	}
119   	
120   	int make_chain(coord ** V, int n, int (*cmp) (const void *, const void *))
121   	{
122   	    int i, j, s = 1;
123   	    coord *t;
124   	
125   	    qsort(V, n, sizeof(coord *), cmp);
126   	    for(i = 2; i < n; i++) {
127   	        for(j = s; j >= 1 && ccw(V, i, j, j - 1); j--)
128   	            ;
129   	        s = j + 1;
130   	        t = V[s];
131   	        V[s] = V[i];
132   	        V[i] = t;
133   	    }
134   	    return s;
135   	}
136   	
137   	int ch2d(coord ** P, int n)
138   	{
139   	    int u = make_chain(P, n, cmpl);    /* make lower hull */
140   	    if(!n)
141   	        return 0;
142   	    P[n] = P[0];
143   	    return u + make_chain(P + u, n - u + 1, cmph);    /* make upper hull */
144   	}
145   	
146   	int plotit(int n, int m, int v[], int art, hpg_t *image, double CHx, double CHy, double CPx, double CPy)
147   	{
148   	    int i;
149   	    int x, y, xl, yl, xf, yf;
150   	    float w = bb_r - bb_l;
151   	    float h = bb_t - bb_b;
152   	    float scale;
153   	
154   	    xl = yl = xf = yf = 0;
155   	
156   	    hpg_clear_on(image);
157   	    hpg_set_color(image,HPG_COLOR_GRAY_13);
158   	
159   	    if(w / h > (float)W / (float)H)
160   	        scale = (W-1) / w;
161   	    else
162   	        scale = (H-1) / h;
163   	
164   	    for(i = 0; i < n; i++) {
165   	        if(art == 0) {
166   	            x = (points[i][0] - bb_l) * scale;
167   	            x += ((W-1) - w * scale) / 2;
168   	            y = (points[i][1] - bb_b) * scale;
169   	            y += ((H-1) - h * scale) / 2;
170   	        } else {
171   	            x = ((points[i][0] - bb_l) / w) * (W-1);
172   	            y = ((points[i][1] - bb_b) / h) * (H-1);
173   	        }
174   	        y = (H-1) - y;
175   	
176   	        hpg_draw_line_on(image,x+1, y, x-1, y);
177   	        hpg_draw_line_on(image,x, y+1, x, y-1);
178   	    }
179   	
180   	    for(i = 0; i < m; i++) {
181   	        if(art == 0) {
182   	            x = (points[v[i]][0] - bb_l) * scale;
183   	            x += ((W-1) - w * scale) / 2;
184   	            y = (points[v[i]][1] - bb_b) * scale;
185   	            y += ((H-1) - h * scale) / 2;
186   	        } else {
187   	            x = ((points[v[i]][0] - bb_l) / w) * (W-1);
188   	            y = ((points[v[i]][1] - bb_b) / h) * (H-1);
189   	        }
190   	        y = (H-1) - y;
191   	
192   	        if (i == 0) {
193   	            xf = xl = x;
194   	            yf = yl = y;
195   	            continue;
196   	        }
197   	
198   	        hpg_set_color(image,HPG_COLOR_GRAY_14);
199   	        hpg_draw_line_on(image,xl, yl, x, y);
200   	
201   	        xl = x;
202   	        yl = y;
203   	    }
204   	    hpg_draw_line_on(image,xl, yl, xf, yf);
205   	
206   	    hpg_set_color(image,HPG_COLOR_BLACK);
207   	    if(art == 0) {
208   	        x = (CHx - bb_l) * scale;
209   	        x += ((W-1) - w * scale) / 2;
210   	        y = (CHy - bb_b) * scale;
211   	        y += ((H-1) - h * scale) / 2;
212   	    } else {
213   	        x = ((CHx - bb_l) / w) * (W-1);
214   	        y = ((CHy - bb_b) / h) * (H-1);
215   	    }
216   	    y = (H-1) - y;
217   	    hpg_fill_circle_on(image,x,y,3);
218   	
219   	    hpg_set_color(image,HPG_COLOR_GRAY_12);
220   	    if(art == 0) {
221   	        x = (CPx - bb_l) * scale;
222   	        x += ((W-1) - w * scale) / 2;
223   	        y = (CPy - bb_b) * scale;
224   	        y += ((H-1) - h * scale) / 2;
225   	    } else {
226   	        x = ((CPx - bb_l) / w) * (W-1);
227   	        y = ((CPy - bb_b) / h) * (H-1);
228   	    }
229   	    y = (H-1) - y;
230   	    hpg_fill_rect_on(image,x-2,y-2,x+2,y+2);
231   	
232   	    return (0);
233   	}
234   	
235   	int grobshot(char *filename, hpg_t *image)
236   	{
237   	    int g, t, x, y;
238   	    unsigned char header[10], p;
239   	    unsigned long int length;
240   	    FILE *fp;
241   	
242   	    if((fp = fopen(filename,"wb")) == NULL) {
243   	        return(1);
244   	    }
245   	
246   	    fwrite("HPHP49-C",1,8,fp);
247   	
248   	    header[0] = 0x1e;
249   	    header[1] = 0x2b;
250   	
251   	    length = (int)ceil(W/8.0) * 2 * H + 15;
252   	
253   	    header[2] = (length & 0x0f) << 4;
254   	    header[3] = (length >> 4) & 0xff;
255   	    header[4] = (length >> 12) & 0xff;
256   	
257   	    header[5] = H & 0xff;
258   	    header[6] = (H >> 8) & 0xff;
259   	    header[7] = ((W & 0x0f) << 4) | ((H >> 16) & 0x0f);
260   	    header[8] = (W >> 4) & 0xff;
261   	    header[9] = (W >> 12) & 0xff;
262   	
263   	    fwrite(header,1,10,fp);
264   	
265   	    for(y = 0; y < H; y++) {
266   	        g = 1;
267   	        t = 0;
268   	        for(x = 0; x < W; x++) {
269   	            if((p = hpg_get_pixel(image, x, y)))
270   	                t += g;
271   	
272   	            g *= 2;
273   	
274   	            if(g == 256) {
275   	                fwrite(&t,1,1,fp);
276   	                t = 0;
277   	                g = 1;
278   	            }
279   	        }
280   	        if(g != 1)
281   	            fwrite(&t,1,1,fp);
282   	    }
283   	
284   	    fclose(fp);
285   	    return (0);
286   	}
287   	
288   	int main(int argc, char *argv[])
289   	{
290   	    int n, m, i, j;
291   	    char *hullpoints, *filename, *ofilename, *ofiletype, msg[40];
292   	    double A = 0, CHx = 0, CHy = 0, PM = 0, CPx, CPy;
293   	    int *v, art, grob=0, bmp=0;
294   	#ifndef HPAPINE
295   	    SAT_LIST p;
296   	#endif
297   	    SAT_STACK_ELEMENT e;
298   	    SAT_STACK_DATA d;
299   	    hpg_t *image;
300   	    FILE *fp;
301   	
302   	    art = sat_pop_real();
303   	    if(art != 0 && art != 1) {
304   	        sat_stack_push_string("Aspect ratio type must be 0 or 1");
305   	        sat_push_real(1);
306   	        return(0);
307   	    }
308   	
309   	    H = sat_pop_real();
310   	    if(H < 1) {
311   	        sat_stack_push_string("Height > 0");
312   	        sat_push_real(1);
313   	        return(0);
314   	    }
315   	
316   	    W = sat_pop_real();
317   	    if(W < 1) {
318   	        sat_stack_push_string("Width > 0");
319   	        sat_push_real(1);
320   	        return(0);
321   	    }
322   	
323   	    sat_get_stack_element(1, &e);
324   	    sat_decode_stack_element(&d, &e);
325   	
326   	    if(d.type == SAT_DATA_TYPE_STRING) {
327   	        ofilename = str_unquote(d.sval, '\'');
328   	        sat_stack_drop();
329   	    } else {
330   	        sat_stack_push_string("Missing output filename");
331   	        sat_push_real(1);
332   	        return (0);
333   	    }
334   	
335   	    if(strlen(ofilename) > 4)
336   	        ofiletype = ofilename + strlen(ofilename) - 3;
337   	
338   	    if(strcmp(ofiletype,"GRO") == 0)
339   	        grob = 1;
340   	    if(strcmp(ofiletype,"BMP") == 0)
341   	        bmp = 1;
342   	
343   	    if(grob + bmp != 1) {
344   	        sat_stack_push_string("File ext. must be .GRO or .BMP");
345   	        sat_push_real(1);
346   	        return (0);
347   	    }
348   	
349   	    sat_get_stack_element(1, &e);
350   	    sat_decode_stack_element(&d, &e);
351   	
352   	    if(d.type == SAT_DATA_TYPE_STRING) {
353   	        filename = str_unquote(d.sval, '\'');
354   	        sat_stack_drop();
355   	    } else {
356   	        sat_stack_push_string("Missing input filename");
357   	        sat_push_real(1);
358   	        return (0);
359   	    }
360   	
361   	    sys_slowOff();
362   	
363   	    if((image = hpg_alloc_gray16_image(W,H)) == NULL) {
364   	        sprintf(msg,"Failed to allocate %dx%d image",W,H);
365   	        sat_stack_push_string(msg);
366   	        sat_push_real(1);
367   	        return(0);
368   	    }
369   	
370   	    if((fp = fopen(ofilename,"w")) == NULL) {
371   	        sprintf(msg,"Cannot open: %s",ofilename);
372   	        sat_stack_push_string(msg);
373   	        sat_push_real(1);
374   	        return(0);
375   	    }
376   	    fclose(fp);
377   	
378   	    hpg_clear();
379   	    hpg_set_font(hpg_stdscreen, hpg_get_bigfont());
380   	    hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);
381   	    hpg_draw_text("Computing Hull...", 15, 38);
382   	
383   	    n = read_points(filename);
384   	    if(n == 0) {
385   	        sprintf(msg,"Cannot read: %s",filename);
386   	        sat_stack_push_string(msg);
387   	        sat_push_real(1);
388   	        sys_slowOn();
389   	        return (0);
390   	    }
391   	    if(n < 3) {
392   	        sat_stack_push_string("Number of points must be > 2!");
393   	        sat_push_real(1);
394   	        sys_slowOn();
395   	        return (0);
396   	    }
397   	
398   	    m = ch2d(P, n);
399   	    v = hull_points(P, m);
400   	
401   	    for(i = 0; i < m; i++) {
402   	        double C, Px, Py;
403   	        int j = (i + 1) % m;
404   	        A += points[v[i]][0] * points[v[j]][1];
405   	        A -= points[v[i]][1] * points[v[j]][0];
406   	        C = ((points[v[i]][0] * points[v[j]][1]) -
407   	             (points[v[j]][0] * points[v[i]][1]));
408   	        CHx += (points[v[i]][0] + points[v[j]][0]) * C;
409   	        CHy += (points[v[i]][1] + points[v[j]][1]) * C;
410   	        Px = (points[v[j]][0] - points[v[i]][0]);
411   	        Py = (points[v[j]][1] - points[v[i]][1]);
412   	        PM += sqrt(Px * Px + Py * Py);
413   	    }
414   	    A /= 2;
415   	    CHx /= (6 * A);
416   	    CHy /= (6 * A);
417   	
418   	    CPx = points[0][0];
419   	    CPy = points[0][1];
420   	    for(i = 1; i < n; i++) {
421   	        CPx = (CPx * i + points[i][0]) / (i + 1);
422   	        CPy = (CPy * i + points[i][1]) / (i + 1);
423   	    }
424   	
425   	    plotit(n, m, v, art, image, CHx, CHy, CPx, CPy);
426   	    hpg_clear();
427   	    hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_BLACK);
428   	
429   	    if(grob) {
430   	        hpg_draw_text("Computing GROB...", 15, 38);
431   	        if(grobshot(ofilename, image) == 1) {
432   	            sprintf(msg,"Cannot open: %s",ofilename);
433   	            sat_stack_push_string(msg);
434   	            sat_push_real(1);
435   	            return(0);
436   	        }
437   	    }
438   	
439   	#define R 2
440   	#define G 1
441   	#define B 0
442   	
443   	    if(bmp) {
444   	        int ret, i;
445   	        palette_t mycolors[16];
446   	
447   	        for(i = 0;i < 16;i++)
448   	            for(j = 0;j < 4;j++)
449   	                mycolors[i].rgb[j] = 0x00;
450   	
451   	        //background
452   	        mycolors[0].rgb[B] = 0xff;
453   	        mycolors[0].rgb[G] = 0xff;
454   	        mycolors[0].rgb[R] = 0xff;
455   	
456   	        //green COG(p)
457   	        mycolors[12].rgb[R] = 0x00;
458   	        mycolors[12].rgb[G] = 0xff;
459   	        mycolors[12].rgb[B] = 0x00;
460   	
461   	        //red sites
462   	        mycolors[13].rgb[R] = 0xff;
463   	        mycolors[13].rgb[G] = 0x00;
464   	        mycolors[13].rgb[B] = 0x00;
465   	
466   	        //blue CH
467   	        mycolors[14].rgb[R] = 0x00;
468   	        mycolors[14].rgb[G] = 0x00;
469   	        mycolors[14].rgb[B] = 0xff;
470   	
471   	        //black COG(h)
472   	        mycolors[15].rgb[R] = 0x00;
473   	        mycolors[15].rgb[G] = 0x00;
474   	        mycolors[15].rgb[B] = 0x00;
475   	
476   	        hpg_draw_text("Computing BMP...", 15, 38);
477   	        if((ret = hpg_save_anybmp_gray16(ofilename,mycolors,image,W,H)) != C_NoError) {
478   	            if(ret == C_ErrOpenFile)
479   	                sprintf(msg,"Cannot open: %s",ofilename);
480   	            if(ret == C_ErrMemory)
481   	                sprintf(msg,"Out of memory");
482   	            if(ret == C_ErrWritePalette)
483   	                sprintf(msg,"Unable to write palette");
484   	
485   	            sat_stack_push_string(msg);
486   	            sat_push_real(1);
487   	            return(0);
488   	        }
489   	    }
490   	
491   	    if((hullpoints = malloc((50 * m + 4) * sizeof(char))) == NULL) {
492   	        sat_stack_push_string("Failed to allocate memory for hull points.");
493   	        sat_push_real(1);
494   	        return (0);
495   	    }
496   	    hullpoints[0] = '\0';
497   	    strcat(hullpoints, "{ ");
498   	    for(i = 0; i < m; i++) {
499   	        char point[50];
500   	        sprintf(point, "{ %0.12E %0.12E } ", points[v[i]][0],
501   	                points[v[i]][1]);
502   	        strcat(hullpoints, point);
503   	    }
504   	    strcat(hullpoints, " }");
505   	
506   	    sat_stack_push_string(hullpoints);
507   	    sat_push_zint_llong(n);
508   	    sat_stack_push_string("n");
509   	    sat_push_real(A);
510   	    sat_stack_push_string("Area");
511   	    sat_push_real(PM);
512   	    sat_stack_push_string("Perimeter");
513   	#ifndef HPAPINE
514   	    p = sat_list_create(20);
515   	    sat_list_add_real(p, CPx);
516   	    sat_list_add_real(p, CPy);
517   	    sat_push_list(p);
518   	    sat_list_destroy(p);
519   	    sat_stack_push_string("COG(p)");
520   	    p = sat_list_create(20);
521   	    sat_list_add_real(p, CHx);
522   	    sat_list_add_real(p, CHy);
523   	    sat_push_list(p);
524   	    sat_list_destroy(p);
525   	    sat_stack_push_string("COG(h)");
526   	#endif
527   	    sat_push_real(0);
528   	    hpg_set_indicator(HPG_INDICATOR_WAIT,HPG_COLOR_WHITE);
529   	    sys_waitRTCTicks(1);
530   	    sys_slowOn();
531   	
532   	    return (0);
533   	}
