#ifndef HPG_BMP_H
#define HPG_BMP_H

#include <hpgraphics.h>

/*!
 * \mainpage  Bmp handling functions
 * Complements to hpg_ functions for handling bmp files.
 */

 /*!
 * \file hpg_bmp.h
 */
 
 /*!
  * \brief No error code
  *
  */
#define C_NoError			0		// No error
 
 /*!
  * \brief Error code when unable to open a file
  *
  */
#define C_ErrOpenFile	1		// Unable to open the file
 
/*!
 * \brief Error code when unable read the header of a file
 *
 */
#define C_ErrReadHeader	2		// Unable to read the header

/*!
 * \brief Error code when the file to open is not a bitmap file
 *
 */
#define C_ErrBadFormat	3		// Not a bitmap file

/*!
 * \brief Error code for not implemented feature (e.g. compression)
 *
 */
#define C_ErrNotImpl		4		// Feature not implemented

/*!
 * \brief Error code when unable to allocate memory
 *
 */
#define C_ErrMemory		5		// Not enough memory

/*!
 * \brief Error code when unable to read the image from the BMP file
 *
 */
#define C_ErrReadImage	6		// Error while reading the image

/*!
 * \brief Error code when unable to write the palette to the BMP file
 *
 */
#define C_ErrWritePalette 7
/*!
 * \brief Structure for storing an element of a palette
 *
 */
typedef struct
{
	unsigned char rgb[4];
} palette_t;
 
/*!
 * \brief Loads a BMP file to a 16-color grayscale image.
 *
 * The resulting image will be the same size as the original pixmap, but will
 * be converted to 16-color gray.
 *
 * ::hpg_free_image should be called when the application is no longer
 * using the image.
 *
 * \param name The name of the bmp file
 * \param ret The return code
 * \return Pointer to an ::hpg_t representing the image.
 */
extern hpg_t *hpg_load_bmp_gray16 (char *name, int *ret);

/*!
 * \brief Saves a 16-color grayscale image to a BMP file (4 bits coded colors).
 *
 * A palette can be passed as parameter, so that the resulting BMP file can be colored.
 *
 * \param name The name of the bmp file
 * \param pal Palette used for the color coding. If NULL, a default palette (levels of gray) is set.
 * \return Pointer to an ::hpg_t representing the image.
 */
extern int hpg_save_bmp_gray16 (char *name, palette_t pal[]);
extern int hpg_save_anybmp_gray16 (char *name, palette_t pal[], hpg_t*, int, int);

/*!
 * \brief Gets some information from a bmp file
 *
 * Gets the size and depth of a bmp file.
 *
 * \param name The name of the bmp file
 * \param width Width of the image
 * \param height Height of the image
 * \param nbcolors Number of colors of the image
 * \return Pointer to an ::hpg_t representing the image.
 */
extern int hpg_get_bmp_info (char *name, int *width, int *height, int *nbcolors);

/*!
 * \brief Gets an error message associated with an error code returne par an hpg_..._bmp_... function
 *
 * \param code The error code
 * \return Pointer to a string
 */
extern char *hpg_bmp_error (int code);
 
#endif
